# Copyright (c) 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class InvalidDriver(Exception):
    """A driver was not found or loaded."""


class PatternNotFound(Exception):
    """A string did not match the expected pattern or regex."""


class InvalidAction(Exception):
    """Raised when attempted a non existent action."""


class ConfigurationError(Exception):
    """An invalid value was used for a Zaqar configuration option."""
